#define MAXPOINT 20000

/*
 * A Box structure. Note that this is a X-ish Box, where x2 and y2 are
 * OUTSIDE the box
 */
typedef struct BoxRecStruct {
    int x1;
    int y1;
    int x2;		/* xmax + 1 */
    int y2;		/* ymax + 1 */
} BoxRec;

typedef void (*Procedure)();

typedef struct ProcStruct {
    char *name;
    Procedure proc;
} Proc;

typedef struct ProcListStruct {
    Proc proc;
    struct ProcListStruct *next;
} ProcList;

typedef struct ParticleStruct {
    int x, y;
    int dx, dy;
    int timeout;
    struct ParticleStruct *next;
} Particle;

typedef struct SystemStruct {
    int ddx, ddy, friction;
    BoxRec pos;
    struct {
	int x1, y1, x2, y2;
	int r1, r2, t1, t2;
    } vel;
    int time_min, time_max;
    float bounce;
    int start_n, start_d, start_c;
    BoxRec clip;
    int colour;
    GC gc;
    ProcList *procs;
    Particle *babies;
    Particle *adults;
    Particle *corpses;
    struct SystemStruct *new_system;
    struct SystemStruct *baby_system;
    struct SystemStruct *next;

    XPoint points1[MAXPOINT];
    XPoint points2[MAXPOINT];
    XPoint *fore_points;
    XPoint *back_points;
    int n_back;
} System;

typedef struct SystemListStruct {
    char name[100];
    System *system;
    struct SystemListStruct *next;
} SystemList;

typedef struct SystemTypeStruct {
    char *systems;
} SystemType;

typedef struct AnimationStruct {
    System *systems;
    Display *dpy;
    Window win;
    GC gc_back;
    Dimension win_height;
} Animation;

#define INT(n)		((n)>>8)
#define FRAC(n)		((n)<<8)

#define RRANGE(lo,hi)	((((random()>>8)&0xFFFF)*((hi)-(lo))>>16) + (lo))

extern Animation *InitAnimation(
#if NeedFunctionPrototypes
System *initial_system
#endif
);

extern void DoAnimation(
#if NeedFunctionPrototypes
Animation *
#endif
);

extern ProcList *CreateProcList();

extern void AddToProcList(
#if NeedFunctionPrototypes
ProcList	*list,
char		*name,
Procedure	proc
#endif
);

extern void AddParticleProcs(
#if NeedFunctionPrototypes
ProcList	*list
#endif
);
